;
; CL.DEF - Default .def file for C Windows applications
;
; Copyright (C) Microsoft Corp 1991.  All rights reserved.
;

NAME           WINDOWAPI

;
; The NAME statement tells the linker that a Windows application is being
; built. The linker supplies the actual name of the application from the
; base name of the executable file.
;

EXETYPE        WINDOWS 3.0

;
; The EXETYPE statement tells the linker to build a Windows 3.0 executable
; file.  This statement should be used in all applications built with
; C 7.0 for Windows 3.0.
;

PROTMODE

;
; The PROTMODE statement tells the linker to mark the application for execution
; in Windows standard or enhanced mode.
;

CODE           PRELOAD MOVEABLE DISCARDABLE
DATA           PRELOAD MOVEABLE

;
; The CODE and DATA statements set the attributes of the applications CODE and
; DATA segments.
;

HEAPSIZE       1024
STACKSIZE      8096

;
; The HEAPSIZE and STACKSIZE statements set the applications near heap and
; stack sizes.  The values specified are recomended for QuickWin applications.
;

;
; Uncomment these lines for DLL module definition file model.
;
;EXPORTS
; WEP @1 RESIDENTNAME
;
;SEGMENTS 'WEP_TEXT' FIXED PRELOAD
;
;
; The above section should be uncommented if this .DEF file is to be used as
; a model for a C 7.0 dynamic-link library (DLL) .DEF file.  The WEP
; routine is included in the C 7.0 startup code, but the symbol must
; be explicitly exported. If a DLL already has a user-defined WEP routine,
; the routine should be renamed _WEP.  The startup code will call _WEP,
; if present, during DLL termination. For additional information concerning
; the WEP function, see the C 7.0 and Windows 3.0 SDK documentation.
;
