s/M4_CHARSTAR_4M/CHARSTAR/g
s/M4_DOUBLESTAR_4M/DOUBLESTAR/g
s/M4_FLOATSTAR_4M/FLOATSTAR/g
s/M4_INTEGERSTAR_4M/INTEGERSTAR/g
s/M4_INTSTAR_4M/INTSTAR/g
s/M4_NAMEF_4M/NAMEF/g
s/M4_STRINGD_4M/STRINGD/g
s/M4_STRINGF_4M/STRINGF/g
s/M4_STRINGL_4M/STRINGL/g
s/M4_STRINGX_4M/STRINGX/g
s/M4_VOIDSTAR_4M/VOIDSTAR/g
s/M4_changecom_4M/changecom/g
s/M4_changequote_4M/changequote/g
s/M4_decr_4M/decr/g
s/M4_define_4M/define/g
s/M4_defn_4M/defn/g
s/M4_divert_4M/divert/g
s/M4_divnum_4M/divnum/g
s/M4_dnl_4M/dnl/g
s/M4_dumpdef_4M/dumpdef/g
s/M4_errprint_4M/errprint/g
s/M4_eval_4M/eval/g
s/M4_format_4M/format/g
s/M4_ifdef_4M/ifdef/g
s/M4_ifelse_4M/ifelse/g
s/M4_include_4M/include/g
s/M4_incr_4M/incr/g
s/M4_index_4M/index/g
s/M4_len_4M/len/g
s/M4_m4exit_4M/m4exit/g
s/M4_m4wrap_4M/m4wrap/g
s/M4_maketemp_4M/maketemp/g
s/M4_popdef_4M/popdef/g
s/M4_pushdef_4M/pushdef/g
s/M4_shift_4M/shift/g
s/M4_sinclude_4M/sinclude/g
s/M4_substr_4M/substr/g
s/M4_syscmd_4M/syscmd/g
s/M4_sysval_4M/sysval/g
s/M4_traceoff_4M/traceoff/g
s/M4_traceon_4M/traceon/g
s/M4_translit_4M/translit/g
s/M4_undefine_4M/undefine/g
s/M4_undivert_4M/undivert/g
s/M4_unix_4M/unix/g
