/****************************************************************************
 * NCSA HDF                                                                 *
 * Software Development Group                                               *
 * National Center for Supercomputing Applications                          *
 * University of Illinois at Urbana-Champaign                               *
 * 605 E. Springfield, Champaign IL 61820                                   *
 *                                                                          *
 * For conditions of distribution and use, see the accompanying             *
 * hdf/COPYING file.                                                        *
 *                                                                          *
 ****************************************************************************/

/*****************************************************************************
 * File: fmpsun.h
 *
 *****************************************************************************/ 

/* $Id: fmplinux.h,v 1.2 1995/09/25 16:16:49 georgev Exp $ */

#ifndef _FMPCONF_H
#define _FMPCONF_H

#define HAVE_PROTOTYPES
#define HAVE_UCHAR
#define HAVE_UINT
#define HAVE_ULONG
#define HAVE_USHORT
#define HAVE_SIZE_T
#define HAVE_SSIZE_T
#define HAVE_SIGSET_T
#define HAVE_LIMITS_H
#define HAVE_UNISTD_H
#define HAVE_STDLIB
#define HAVE_STDIO_H
#define HAVE_FCNTL
#define HAVE_STAT
#define HAVE_MIN_MAX
#define HAVE_CDEFS_H
#define	BYTE_ORDER  LITTLE_ENDIAN	

#endif /* _FMPCONF_H */




